"""
Simple TUI editor
"""

# Original code:
# https://github.com/pfalcon/picotui/blob/master/seditor.py
#
# Very simple VT100 terminal text editor widget
# Copyright (c) 2015 Paul Sokolovsky, (c) 2022 Albert Zeyer
# Distributed under MIT License

# https://en.wikipedia.org/wiki/ANSI_escape_code
# https://invisible-island.net/xterm/ctlseqs/ctlseqs.html

from __future__ import annotations
from typing import Optional, Union
from .tty_helpers import *


class TuiEditor:
    """TUI editor"""

    def __init__(self, *,
                 control_key_abort: Optional[str] = "C",  # raises KeyboardInterrupt
                 control_key_quit: Optional[str] = "S",  # normal quit
                 ):
        self.top_line_idx = 0
        self.row = 0
        self.col = 0
        self.height = 10  # lines for editor only, excluding status bars
        self.show_line_numbers = False
        self._orig_termios = None
        self._orig_sig_win_ch = None
        self.content_prefix_escape = b"\x1b[30;106m"
        self.status_prefix_escape = b"\x1b[30;102m"
        self._content = [""]
        self._status_content = [""]
        self._control_key_abort = _control_key(control_key_abort)
        self._control_key_quit = _control_key(control_key_quit)
        self.tty = TtyController(
            total_height=lambda: self.actual_height + len(self._status_content),
            update_screen=self.update_screen,
        )

    def edit(self):
        """enter raw tty, enter loop, and exit raw tty at the end"""
        self.tty.init_tty()
        try:
            self.update_screen()
            self.loop()
        finally:
            self.tty.deinit_tty()

    def set_text_lines(self, lines: list[str]):
        """set editor text"""
        self._content = lines or [""]

    def set_text(self, text: str):
        """set editor text"""
        self._content = text.split("\n")

    def get_text_lines(self):
        """get editor text"""
        return self._content

    def get_text(self) -> str:
        """get editor text"""
        return "\n".join(self._content)

    def set_status_lines(self, lines: list[str]):
        """set status bar"""
        self._status_content = lines or [""]
        self.update_screen_status()

    def set_cursor_pos(self, line: int, col: int = 0):
        """
        Set the cursor position (row/col), by line/col
        """
        if self.top_line_idx > line or self.top_line_idx + self.actual_height <= line:
            self.top_line_idx = max(line - self.actual_height // 2, 0)
        self.row = line - self.top_line_idx
        self.col = col

    def set_cursor(self):
        """set the cursor back to the editor pos"""
        width = self.tty.width
        col = self.col
        if self.show_line_numbers:
            col += len(str(self.total_lines)) + 2
        while col >= width - 1:
            col -= width // 3  # see _show_content_line
        self.tty.goto(self.row, col)

    def adjust_cursor_eol(self):
        """when the cur line changed, potentially fix cursor pos col"""
        cur_line_len = len(self._content[self.cur_line_idx])
        if self.col > cur_line_len:
            self.col = cur_line_len

    @property
    def total_lines(self):
        """total number of text lines (potentially not all are visible)"""
        return len(self._content)

    @property
    def cur_line_idx(self) -> int:
        """cursor line index"""
        return self.top_line_idx + self.row

    def loop(self):
        """main loop, reading user inputs"""
        while True:
            buf = os.read(self.tty.fd_in, 32)
            if buf and buf[:1] not in b"\x1b\x7f" and buf[0] > 31:
                try:
                    buf = buf.decode("utf8")
                except UnicodeDecodeError:
                    pass
            sz = len(buf)
            i = 0
            while i < sz:
                if isinstance(buf, bytes) and buf[0] == 0x1b:
                    key = buf
                    i = len(buf)
                else:
                    key = buf[i:i + 1]
                    i += 1
                if key in KEYMAP:
                    key = KEYMAP[key]
                if self.on_key(key):
                    continue
                if key == self._control_key_abort:
                    raise KeyboardInterrupt
                if key == self._control_key_quit:
                    return
                if self.handle_cursor_keys(key):
                    self.on_cursor_pos_change()
                    continue
                self.handle_key(key)
                self.on_cursor_pos_change()

    def update_screen(self):
        """update the screen, i.e. editor and status bar(s)"""
        self.tty.cursor(False)
        self.tty.goto(0, 0)
        self.tty.write(self.content_prefix_escape)
        i = self.top_line_idx
        for c in range(self.height):
            self._show_content_line(i)
            self.tty.write(b"\r\n")
            i += 1
            if i == self.total_lines:
                break
        self.update_screen_status(goto=False)
        self.set_cursor()
        self.tty.cursor(True)

    @property
    def actual_height(self) -> int:
        """actual height of editor (excluding status bars)"""
        return min(self.height, self.total_lines)

    def update_screen_status(self, *, goto=True):
        """
        Update the screen status bar(s).

        :param goto: set the cursor ot the status bar, and then back to the editor
        """
        if goto:
            self.tty.cursor(False)
            self.tty.goto(self.actual_height, 0)
            self.tty.update_occupied_space()
        self.tty.write(self.status_prefix_escape)
        assert self._status_content
        for c, line in enumerate(self._status_content):
            if c > 0:
                self.tty.write(b"\n")
            self._show_line(line)
            self.tty.write(b"\r")
        self.tty.write(b"\x1b[0m")
        if goto:
            self.set_cursor()
            self.tty.cursor(True)

    def update_line(self, *, margin_logic_cur_line=True):
        """
        Update just the current line, assuming that the cursor is in the right line.
        """
        self.tty.cursor(False)
        self.tty.write(b"\r")
        self.tty.write(self.content_prefix_escape)
        self._show_content_line(self.cur_line_idx, margin_logic_cur_line=margin_logic_cur_line)
        self.set_cursor()
        self.tty.cursor(True)

    def _maybe_update_line_after_cursor_change(self, *, is_old_line=False):
        self.set_cursor()
        line = self._content[self.cur_line_idx]
        width = self.tty.width
        if self.show_line_numbers:
            width -= len(str(self.total_lines)) + 2
        if len(line) >= width:
            self.update_line(margin_logic_cur_line=not is_old_line)
            return True
        return False

    def _show_content_line(self, line_idx: int, *, margin_logic_cur_line=True):
        self._show_line_number(line_idx)
        line = self._content[line_idx]
        if margin_logic_cur_line and line_idx == self.cur_line_idx:
            width = self.tty.width
            col = self.col
            if self.show_line_numbers:
                col += len(str(self.total_lines)) + 2
            while col >= width - 1:  # keep consistent to set_cursor
                col -= width // 3
                line = line[width // 3:]
        self._show_line(line)

    def _show_line_number(self, line_idx: int = None):
        if not self.show_line_numbers:
            return
        if line_idx is None:
            line_idx = self.cur_line_idx
        s = str(line_idx + 1)
        s = " " * (len(str(self.total_lines)) - len(s)) + s + "| "
        self.tty.write(s.encode("utf8"))

    def _show_line(self, line: str):
        self.tty.write(line.encode("utf8"))
        self.tty.clear_to_eol()

    def _move_cursor_next_line(self, *, update=True):
        self._maybe_update_line_after_cursor_change(is_old_line=True)
        if self.row + 1 == self.height:
            self.top_line_idx += 1
            self.adjust_cursor_eol()
            if update:
                self.update_screen()
            return True
        else:
            self.row += 1
            self.adjust_cursor_eol()
            if update:
                self._maybe_update_line_after_cursor_change()
            return False

    def _move_cursor_prev_line(self, *, update=True):
        self._maybe_update_line_after_cursor_change(is_old_line=True)
        if self.row == 0:
            if self.top_line_idx > 0:
                self.top_line_idx -= 1
                self.adjust_cursor_eol()
                if update:
                    self.update_screen()
                return True
            return False
        else:
            self.row -= 1
            self.adjust_cursor_eol()
            if update:
                self._maybe_update_line_after_cursor_change()
            return False

    def handle_cursor_keys(self, key):
        """
        Handle potential cursor-position-manipulating key,
        potentially updating the cursor position
        and potentially updating the screen.

        :return: True, if we consumed the event, and it should not be handled further
        """
        if key == KEY_DOWN:
            if self.cur_line_idx + 1 != self.total_lines:
                self._move_cursor_next_line()
        elif key == KEY_UP:
            if self.cur_line_idx > 0:
                self._move_cursor_prev_line()
        elif key == KEY_LEFT:
            if self.col > 0:
                self.col -= 1
                self._maybe_update_line_after_cursor_change()
            elif self.cur_line_idx > 0:
                self.col = len(self._content[self.cur_line_idx - 1])
                self._move_cursor_prev_line()
        elif key == KEY_RIGHT:
            if self.col < len(self._content[self.cur_line_idx]):
                self.col += 1
                self._maybe_update_line_after_cursor_change()
            elif self.cur_line_idx < len(self._content) - 1:
                self.col = 0
                self._move_cursor_next_line()
        elif key == KEY_HOME:
            self.col = 0
            self._maybe_update_line_after_cursor_change()
        elif key == KEY_END:
            self.col = len(self._content[self.cur_line_idx])
            self._maybe_update_line_after_cursor_change()
        elif key == KEY_PGUP:
            self.top_line_idx -= self.height
            if self.top_line_idx < 0:
                self.top_line_idx = 0
                self.row = 0
            elif self.cur_line_idx < 0:
                self.row = 0
            self.adjust_cursor_eol()
            self.update_screen()
        elif key == KEY_PGDN:
            self.top_line_idx += self.height
            if self.cur_line_idx >= self.total_lines:
                self.top_line_idx = self.total_lines - self.height
                if self.top_line_idx >= 0:
                    self.row = self.height - 1
                else:
                    self.top_line_idx = 0
                    self.row = self.cur_line_idx
            self.adjust_cursor_eol()
            self.update_screen()
        else:
            return False
        return True

    def handle_key(self, key: Union[bytes, str, int]):
        """
        Handle potential edit key and update screen.
        Assumes the cursor is at the right position.
        """
        cur_line = self._content[self.cur_line_idx]
        if key == KEY_ENTER:
            self._content[self.cur_line_idx] = cur_line[:self.col]
            self._content.insert(self.cur_line_idx + 1, cur_line[self.col:])
            self.col = 0
            self._move_cursor_next_line(update=False)
            self.tty.update_occupied_space()  # actual height will increase
            self.update_screen()
        elif key == KEY_BACKSPACE:
            if self.col > 0:
                self.col -= 1
                cur_line = cur_line[:self.col] + cur_line[self.col + 1:]
                self._content[self.cur_line_idx] = cur_line
                self.update_line()
            elif self.col == 0 and self.cur_line_idx > 0:
                self.col = len(self._content[self.cur_line_idx - 1])
                self._content[self.cur_line_idx - 1] += self._content[self.cur_line_idx]
                self._content.pop(self.cur_line_idx)
                if self.top_line_idx > 0 and self.top_line_idx + self.height > len(self._content):
                    self.top_line_idx -= 1
                elif self.row > 0:
                    self.row -= 1
                self.tty.update_occupied_space()  # actual height might have decreased
                self.update_screen()
        elif key == KEY_DELETE:
            if self.col < len(cur_line):
                cur_line = cur_line[:self.col] + cur_line[self.col + 1:]
                self._content[self.cur_line_idx] = cur_line
                self.update_line()
            elif self.col == len(cur_line) and self.cur_line_idx < len(self._content) - 1:
                self._content[self.cur_line_idx] += self._content[self.cur_line_idx + 1]
                self._content.pop(self.cur_line_idx + 1)
                self.tty.update_occupied_space()  # actual height might have decreased
                self.update_screen()
        elif isinstance(key, int):
            return
        elif isinstance(key, bytes) and key[0] == 0x1b:
            return
        elif len(key) == 1 and ord(key) <= 31:  # other control char
            return
        elif isinstance(key, str):
            cur_line = cur_line[:self.col] + key + cur_line[self.col:]
            self._content[self.cur_line_idx] = cur_line
            self.col += 1
            self.update_line()
        self.on_edit()

    def on_key(self, key: Union[bytes, str, int]) -> Union[bool, None]:
        """
        Called on input key events.

        :param key:
        :return: True -> the function consumed the event, i.e. the event should not be passed on further
        """
        # Overwrite this function if you want.
        pass

    def on_cursor_pos_change(self):
        """
        Called on (potential) cursor position changes.
        """
        # Overwrite this function if you want.
        pass

    def on_edit(self):
        """
        Called on (potential) edits.
        """
        # Overwrite this function if you want.
        pass


def _control_key(key: Optional[str]) -> Optional[bytes]:
    if not key:
        return None
    key = key.upper()
    assert ord("A") <= ord(key) <= ord("Z"), f"invalid control key {key!r}, should be in range A-Z"
    return bytes([ord(key) - ord("A") + 1])



"""
TTY helpers
"""

# https://en.wikipedia.org/wiki/ANSI_escape_code
# https://invisible-island.net/xterm/ctlseqs/ctlseqs.html

from __future__ import annotations
from typing import Callable
import os
import re
import tty
import termios
import signal


class TtyController:
    """TTY controller"""

    def __init__(self, *,
                 fd_in: int = 0,
                 fd_out: int = 1,
                 total_height: Callable[[], int],  # for reserving space
                 update_screen: Callable[[], None],  # called on resize
                 ):
        self.fd_in = fd_in
        self.fd_out = fd_out
        self.row = 0
        self.col = 0
        self.screen_top = 0
        self.width = os.get_terminal_size(self.fd_out).columns
        self.prev_total_height = 1  # assume we are already in a new line
        self.total_height = total_height
        self.update_screen = update_screen
        self._orig_termios = None
        self._orig_sig_win_ch = None

    def init_tty(self):
        """init raw TTY mode, reserve space (via total_height)"""
        self._orig_termios = termios.tcgetattr(self.fd_in)
        tty.setraw(self.fd_in)
        self.write(b"\x1b[?7l")  # No Auto-Wrap Mode (DECAWM)

        # make enough space
        # assuming nothing has been printed yet
        self.update_occupied_space()

        def _on_resize(_signum, _frame):
            self._update_editor_row_offset()
            self.width = os.get_terminal_size(self.fd_out).columns
            # If the colum size changed, and this wraps around existing text,
            # this is not handled correctly yet...
            # Updating the screen might be a good idea anyway.
            self.update_screen()

        self._orig_sig_win_ch = signal.getsignal(signal.SIGWINCH)
        signal.signal(signal.SIGWINCH, _on_resize)

    def deinit_tty(self, clear_editor=True):
        """resets the tty to normal (non-raw) mode, recovering the original state before init_tty"""
        self.write(b"\x1b[0m")
        self.write(b"\x1b[?7h")  # Auto-Wrap Mode (DECAWM)
        num_lines = self.total_height()
        if clear_editor:
            self.goto(0, 0)
            self.write(b"\x1b[%iM" % num_lines)
        else:
            # Don't leave cursor in the middle of screen
            self.goto(num_lines - 1, 0)
            self.write(b"\r\n")
        termios.tcsetattr(self.fd_in, termios.TCSANOW, self._orig_termios)
        signal.signal(signal.SIGWINCH, self._orig_sig_win_ch)

    def write(self, s: bytes):
        """raw write"""
        assert isinstance(s, bytes)
        os.write(self.fd_out, s)

    def cls(self):
        """clear screen"""
        self.write(b"\x1b[2J")

    def goto(self, row: int, col: int):
        """change cursor position (relative to screen top)"""
        self.write(b"\x1b[%d;%dH" % (row + 1 + self.screen_top, col + 1))
        self.row = row
        self.col = col

    def get_cursor_pos_abs(self) -> (int, int):
        """get absolute cursor position"""
        self.write(b"\x1b[6n")
        s = b""
        while True:
            s += os.read(self.fd_in, 1)
            if s[-1:] == b"R":
                break
            if s[-1:] == b"\x03":
                raise KeyboardInterrupt
        res = re.match(rb".*\[(?P<y>\d*);(?P<x>\d*)R", s)
        row, col = res.groups()
        return int(row) - 1, int(col) - 1

    def clear_to_eol(self):
        """clear to end of line"""
        self.write(b"\x1b[0K")

    def cursor(self, enabled: bool):
        """show or hide cursor"""
        if enabled:
            self.write(b"\x1b[?25h")
        else:
            self.write(b"\x1b[?25l")

    def _update_editor_row_offset(self):
        row, col = self.get_cursor_pos_abs()
        expected_row = self.screen_top + self.row
        self.screen_top += row - expected_row

    def update_occupied_space(self):
        """
        Update occupied space.
        When new lines are added, they are always added at the end.
        When lines are deleted, they are deleted at the current cursor.
        The cursor row stays the same.
        """
        total_height = self.total_height()
        assert total_height > 0 and self.prev_total_height > 0
        if total_height == self.prev_total_height:
            return

        self.cursor(False)
        self.write(b"\r\x1b[0m")
        prev_row = self.row
        if total_height > self.prev_total_height:
            if self.row != self.prev_total_height - 1:
                self.goto(self.prev_total_height - 1, 0)
            num_lines = total_height - self.prev_total_height
            self.write(b"\n" * num_lines)
            self.row = total_height - 1
            self._update_editor_row_offset()
            self.goto(prev_row, 0)
        elif total_height < self.prev_total_height:
            num_lines = self.prev_total_height - total_height
            self.write(b"\x1b[%iM" % num_lines)  # delete lines
        self.prev_total_height = total_height


KEY_UP = 1
KEY_DOWN = 2
KEY_LEFT = 3
KEY_RIGHT = 4
KEY_HOME = 5
KEY_END = 6
KEY_PGUP = 7
KEY_PGDN = 8
KEY_ENTER = 10
KEY_BACKSPACE = 11
KEY_DELETE = 12

KEYMAP = {
    b"\x1b[A": KEY_UP,
    b"\x1b[B": KEY_DOWN,
    b"\x1b[D": KEY_LEFT,
    b"\x1b[C": KEY_RIGHT,
    b"\x1bOH": KEY_HOME,
    b"\x1bOF": KEY_END,
    b"\x1bb": KEY_HOME,
    b"\x1bf": KEY_END,
    b"\x1b[1~": KEY_HOME,
    b"\x1b[4~": KEY_END,
    b"\x1b[5~": KEY_PGUP,
    b"\x1b[6~": KEY_PGDN,
    b"\r": KEY_ENTER,
    b"\x7f": KEY_BACKSPACE,
    b"\x1b[3~": KEY_DELETE,
}
